// src/backend/scheduler.cc - Copyright 2005, 2006, University
//               of Padova, dept. of Pure and Applied
//               Mathematics
//
// This file is part of SGPEMv2.
//
// This is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// SGPEMv2 is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with SGPEMv2; if not, write to the Free Software
// Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA

#include "concrete_environment.hh"
#include "concrete_history.hh"
#include "policy.hh"
#include "schedulable.hh"
#include "scheduler.hh"
#include "policy_manager.hh"
#include "policies_gatekeeper.hh"
#include "user_interrupt_exception.hh"

// Do not include full template definition in the header file
#include "singleton.tcc"

#include <iostream>
using namespace std;
using namespace sgpem;

// Explicit template instantiation to allow to export symbols from the DSO.
template class SG_DLLEXPORT Singleton<Scheduler>;


// ------------------ Static helper functions --------------


// Collects all threads of an environment into a single vector
static void 
collect_threads(const std::vector<Process*>& procs, 
		std::vector<DynamicThread*>& collected_threads)
{
  typedef std::vector<Process*> Processes;
  typedef std::vector<Thread*>  Threads;

  collected_threads.clear();
  for(Processes::const_iterator it1 = procs.begin(); it1 != procs.end(); it1++)
    {
      const Threads& ts = (*it1)->get_threads();
      for(Threads::const_iterator it2 = ts.begin(); it2 != ts.end(); it2++) 
	collected_threads.push_back((DynamicThread*) *it2);
    }
}


// ---------------------------------------------------------




//private constructor. The parameter is discarded
Scheduler::Scheduler()
    : _policy_manager(PolicyManager::get_registered_manager())
{
  _policy_manager.init();
}

ReadyQueue*
Scheduler::get_ready_queue()
{
  // FIXME return the correct queue accordingly to the value returned by Policy::wants()
  return &_ready_queue;
}


/** \note E' fondamentale che questo metodo memorizzi localmente qualora la politica
 attuale sia a prerilascio o meno, e la durata del quanto di tempo, in quanto la politica
 e' libera di variare questi parametri a piacere durante l'esecuzione della simulazione
*/
void
Scheduler::reset_status()
{
  // DEPRECATED (?)
}


Policy&
Scheduler::get_policy()
{
  // FIXME : fixme.
  // return *PoliciesGatekeeper::get_instance().get_current_policy(&History::get_instance());
}


void
Scheduler::step_forward(History& history, Policy& cpu_policy) throw(UserInterruptException)
{
  // NOTE: Be sure to read the *ORIGINAL* documentation in the design document for this method!

  // FIXME: handle me! I'm not just a pretty boolean, I want to be *USED*! *EXPLOITED*!
  //        *RAPED*! *MAKE ME BLEED*!
  bool simulation_ended = true; // Assume we've finished. Then prove me wrong.

  ConcreteHistory& concrete_history = (ConcreteHistory&) history;
  ConcreteEnvironment* new_snapshot = new ConcreteEnvironment(concrete_history.get_last_environment());
  
  typedef std::vector<DynamicProcess*> Processes;
  typedef std::vector<DynamicThread*>  Threads;
  
  Threads all_threads;

  collect_threads(new_snapshot->get_processes(), all_threads);

  // designer + implementer (Matteo) comment follows:
  // FIXME: decreasing the time elapsed of the running thread + process 
  // should maybe be done here as the first thing, instead than 
  // directly when selecting them

  // 1. mark future threads as ready, if appropriate
  // 2. mark ended threads as terminated
  for(Threads::iterator it = all_threads.begin(); it != all_threads.end(); it++)
    {
      if((*it)->get_state() == Schedulable::state_future)
	{
	  Process& parent = (*it)->get_process();
	  if(parent.get_elapsed_time() == (*it)->get_arrival_time())
	    (*it)->set_state(Schedulable::state_ready);
	}
      
      if((*it)->get_total_cpu_time() - (*it)->get_elapsed_time() == 0)
	(*it)->set_state(Schedulable::state_terminated);
      
      // 3. check for simulation termination (we can directly use threads)
      // for this check, since processes' state is based upon theirs)
      if(simulation_ended && 
	 ((*it)->get_state() & (Schedulable::state_blocked | 
				Schedulable::state_terminated)) == 0)
	simulation_ended = false;
    }


  //                    /
  //                  /
  //                /
  // (I'M HERE)   < * * * * * * * * * * *  
  //		    \
  //                  \
  //                    \
  //
  // (is it visible enough for you?)

  try
  {
    // call the policy to sort the queue
    // ...
  }
  catch( UserInterruptException e )
  {
    _policy_manager.init();

    //TODO Do we need to perform some other cleanup operation here?
    delete new_snapshot;

    // Do we need to update something?

    // Going up unwinding the stack, tell:
    //    - the user that the policy sucks
    //    - SimulationController that everything stopped
    throw;
  }
  
  concrete_history.append_new_environment(new_snapshot);
}
